program lndcake;
uses Windows;


///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
/////////////////////////////// Consts \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
const
 VerData = 'LndCake v1.01';

 OpenLNDTitle: pchar = 'Open LND file';
 LNDFilter: packed array[1..30] of char = (
  'B', '&', 'W', ' ', 'l', 'a', 'n', 'd', ' ', 'f', 'i', 'l', 'e', 's', ' ', '(', '*', '.', 'l', 'n', 'd', ')', #0, '*', '.', 'L', 'N', 'D', #0,
  #0);
 SaveLNDTitle: pchar = 'Save LND file';


///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
////////////////////////////// Typedefs \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
type
//TLndFile = imaginary record
// Header: TLndHeader;
// LoResTexture: array[1..Header.NumLoResTextures] of TLndLoResTex;
// Block: array[1..Header.NumBlocks] of TLndBlock;
// Country: array[1..Header.NumCountries] of TLndCountry;
// Material: array[1..Header.NumMaterials] of TLndMaterial;
// NoiseMap: array[0..255,0..255] of byte;
// BumpMap: array[0..255,0..255] of byte;
//end;

 TLndHeader = packed record
  NumBlocks: longint;
  BlockIndex: array[0..31,0..31] of byte;
  NumMaterials: longint;
  NumCountries: longint;
  BlockSize: longint;    // sizeof(TLndBlock)    =   2520
  MaterialSize: longint; // sizeof(TLndMaterial) = 131074
  CountrySize: longint;  // sizeof(TLndCountry)  =   3076
  NumLoResTextures: longword;
 end;

 TLndLoResTex = packed record
  Texture: pointer; // ignore
  Material: pointer; // ignore
  NumSubTextures: longint;
  ID: longint;
  Size: longint;
  // (Size - 4) bytes of DirectDraw texture crap
 end;

 TLndCell = packed record
  Color: array[1..4] of byte; // or something
  Altitude: byte;
  SaveColor: byte;
  Flags: array[1..2] of byte;
 end;
 TLndBlock = packed record
  Cell: array[0..16,0..16] of TLndCell;
  Index: longint;
  Dummy: array[1..204] of byte;
 end;


///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//////////////////////////////// Vars \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
var
 LND: HANDLE;
 LNDOFN: OPENFILENAME;
 FNbuf, FNbuf2: array[1..320] of char;
 Header: TLndHeader;
 LoResTexture: TLndLoResTex;
 Block: TLndBlock;
 X, Y, Z, BytesRead, BytesWritten: longword;
 i, j: word;


///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//////////////////////////////// Code \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
begin
 writeln(VerData);
 writeln;

 // Skip this if command line gives a filename?
 LNDOFN.lStructSize := sizeof(LNDOFN);
 LNDOFN.lpstrFilter := @LNDFilter;
 LNDOFN.lpstrFile := @FNbuf;
 LNDOFN.nMaxFile := sizeof(FNbuf);
 LNDOFN.lpstrTitle := OpenLNDTitle;
 LNDOFN.Flags := OFN_EXPLORER + OFN_FILEMUSTEXIST + OFN_HIDEREADONLY;
 if GetOpenFileName(@LNDOFN) then
  writeln('Loading ', FNbuf, '...')
 else // We have no LND file to edit.
  exit;

 FNbuf2 := FNbuf;
 LNDOFN.lpstrFile := @FNbuf2;
 LNDOFN.lpstrTitle := SaveLNDTitle;
 LNDOFN.Flags := OFN_EXPLORER + OFN_OVERWRITEPROMPT;
 if GetSaveFileName(@LNDOFN) then
  if not CopyFile(@FNbuf, @FNbuf2, FALSE) then
   begin
    writeln('Error ', GetLastError, ' preparing output file!');
    exit;
   end
 else // We have no LND file to edit.
  exit;
  
 LND := CreateFile(@FNbuf2, (GENERIC_READ + GENERIC_WRITE), FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
 if (LND = INVALID_HANDLE_VALUE) then
  begin
   writeln('Error ', GetLastError, ' opening ', FNbuf2, '!');
   exit;
  end;

 if not ReadFile(LND, Header, SizeOf(Header), BytesRead, nil) then
  begin
   writeln('Error ', GetLastError, ' reading LND header!');
   exit;
  end;

 writeln('Header info:');
 writeln(' NumBlocks        ', Header.NumBlocks);
 writeln(' NumMaterials     ', Header.NumMaterials);
 writeln(' NumCountries     ', Header.NumCountries);
 writeln(' BlockSize        ', Header.BlockSize);
 writeln(' MaterialSize     ', Header.MaterialSize);
 writeln(' CountrySize      ', Header.CountrySize);
 writeln(' NumLoResTextures ', Header.NumLoResTextures);
 writeln;

 for i := 1 to Header.NumLoResTextures do
  begin // skip low-res textures
   ReadFile(LND, LoResTexture, SizeOf(LoResTexture), BytesRead, nil);
   SetFilePointer(LND, (LoResTexture.Size - 4), nil, FILE_CURRENT);
  end;
 // ...OR jump to (filesize-(131072 + Header.NumCountries*Header.CountrySize + Header.NumMaterials*Header.MaterialSize + Header.NumBlocks*Header.BlockSize))

 writeln('Editing cell data:');
 for i := 2 to Header.NumBlocks do // why the hell...?
  begin
   Z := SetFilePointer(LND, 0, nil, FILE_CURRENT);
   ReadFile(LND, Block, SizeOf(Block), BytesRead, nil);
   write(' Block #', Block.Index, '...');
   for Y := 0 to 16 do
    for X := 0 to 16 do
     case Block.Cell[Y,X].Altitude of
      1..4: Block.Cell[Y,X].Altitude := Block.Cell[Y,X].Altitude;
      5..7: Block.Cell[Y,X].Altitude := 4;
      8..10: Block.Cell[Y,X].Altitude := 5;
     else
      Block.Cell[Y,X].Altitude := Block.Cell[Y,X].Altitude shr 1;
     end;

   SetFilePointer(LND, Z, nil, FILE_BEGIN);
   WriteFile(LND, Block, SizeOf(Block), BytesWritten, nil);
   writeln(' done.');
  end;
 writeln;
 writeln('Finished editing.');
 CloseHandle(LND);
end.
